/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "IPAddrAndASCertExtn"
 * 	found in "rfc3779.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "IPAddressRange.h"

asn_TYPE_member_t asn_MBR_IPAddressRange_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct IPAddressRange, min),
		(ASN_TAG_CLASS_UNIVERSAL | (3 << 2)),
		0,
		&asn_DEF_IPAddress,
		0,
		{ 0, 0, 0 },
		0, 0, /* No default value */
		"min"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct IPAddressRange, max),
		(ASN_TAG_CLASS_UNIVERSAL | (3 << 2)),
		0,
		&asn_DEF_IPAddress,
		0,
		{ 0, 0, 0 },
		0, 0, /* No default value */
		"max"
		},
};
static const ber_tlv_tag_t asn_DEF_IPAddressRange_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static const asn_TYPE_tag2member_t asn_MAP_IPAddressRange_tag2el_1[] = {
    { (ASN_TAG_CLASS_UNIVERSAL | (3 << 2)), 0, 0, 1 }, /* min */
    { (ASN_TAG_CLASS_UNIVERSAL | (3 << 2)), 1, -1, 0 } /* max */
};
asn_SEQUENCE_specifics_t asn_SPC_IPAddressRange_specs_1 = {
	sizeof(struct IPAddressRange),
	offsetof(struct IPAddressRange, _asn_ctx),
	asn_MAP_IPAddressRange_tag2el_1,
	2,	/* Count of tags in the map */
	0, 0, 0,	/* Optional elements (not needed) */
	-1,	/* First extension addition */
};
asn_TYPE_descriptor_t asn_DEF_IPAddressRange = {
	"IPAddressRange",
	"IPAddressRange",
	&asn_OP_SEQUENCE,
	asn_DEF_IPAddressRange_tags_1,
	sizeof(asn_DEF_IPAddressRange_tags_1)
		/sizeof(asn_DEF_IPAddressRange_tags_1[0]), /* 1 */
	asn_DEF_IPAddressRange_tags_1,	/* Same as above */
	sizeof(asn_DEF_IPAddressRange_tags_1)
		/sizeof(asn_DEF_IPAddressRange_tags_1[0]), /* 1 */
	{ 0, 0, SEQUENCE_constraint },
	asn_MBR_IPAddressRange_1,
	2,	/* Elements count */
	&asn_SPC_IPAddressRange_specs_1	/* Additional specs */
};

