#ifndef SRC_OBJECT_ROA_H_
#define SRC_OBJECT_ROA_H_

#include "rpp.h"
#include "types/address.h"
#include "types/uri.h"

int roa_traverse(struct rpki_uri *, struct rpp *);

#endif /* SRC_OBJECT_ROA_H_ */
