/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "IPAddrAndASCertExtn"
 * 	found in "rfc3779.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_ASIdOrRange_H_
#define	_ASIdOrRange_H_


#include "asn1/asn1c/asn_application.h"

/* Including external dependencies */
#include "ASId.h"
#include "ASRange.h"
#include "asn1/asn1c/constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum ASIdOrRange_PR {
	ASIdOrRange_PR_NOTHING,	/* No components present */
	ASIdOrRange_PR_id,
	ASIdOrRange_PR_range
} ASIdOrRange_PR;

/* ASIdOrRange */
typedef struct ASIdOrRange {
	ASIdOrRange_PR present;
	union ASIdOrRange_u {
		ASId_t	 id;
		ASRange_t	 range;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} ASIdOrRange_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_ASIdOrRange;
extern asn_CHOICE_specifics_t asn_SPC_ASIdOrRange_specs_1;
extern asn_TYPE_member_t asn_MBR_ASIdOrRange_1[2];

#ifdef __cplusplus
}
#endif

#endif	/* _ASIdOrRange_H_ */
#include "asn1/asn1c/asn_internal.h"
