/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "RPKI-ROA"
 * 	found in "rfc6482.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "ROAIPAddressFamily.h"

static int
memb_addressFamily_constraint_1(const asn_TYPE_descriptor_t *td, const void *sptr,
			asn_app_constraint_failed_f *ctfailcb, void *app_key) {
	const OCTET_STRING_t *st = (const OCTET_STRING_t *)sptr;
	size_t size;
	
	if(!sptr) {
		ASN__CTFAIL(app_key, td, sptr,
			"%s: value not given (%s:%d)",
			td->name, __FILE__, __LINE__);
		return -1;
	}
	
	size = st->size;
	
	if((size >= 2 && size <= 3)) {
		/* Constraint check succeeded */
		return 0;
	} else {
		ASN__CTFAIL(app_key, td, sptr,
			"%s: constraint failed (%s:%d)",
			td->name, __FILE__, __LINE__);
		return -1;
	}
}

static int
memb_addresses_constraint_1(const asn_TYPE_descriptor_t *td, const void *sptr,
			asn_app_constraint_failed_f *ctfailcb, void *app_key) {
	size_t size;
	
	if(!sptr) {
		ASN__CTFAIL(app_key, td, sptr,
			"%s: value not given (%s:%d)",
			td->name, __FILE__, __LINE__);
		return -1;
	}
	
	/* Determine the number of elements */
	size = _A_CSEQUENCE_FROM_VOID(sptr)->count;
	
	if((size >= 1)) {
		/* Perform validation of the inner elements */
		return td->encoding_constraints.general_constraints(td, sptr, ctfailcb, app_key);
	} else {
		ASN__CTFAIL(app_key, td, sptr,
			"%s: constraint failed (%s:%d)",
			td->name, __FILE__, __LINE__);
		return -1;
	}
}

static asn_oer_constraints_t asn_OER_type_addresses_constr_3 CC_NOTUSED = {
	{ 0, 0 },
	-1	/* (SIZE(1..MAX)) */};
static asn_oer_constraints_t asn_OER_memb_addressFamily_constr_2 CC_NOTUSED = {
	{ 0, 0 },
	-1	/* (SIZE(2..3)) */};
static asn_oer_constraints_t asn_OER_memb_addresses_constr_3 CC_NOTUSED = {
	{ 0, 0 },
	-1	/* (SIZE(1..MAX)) */};
static asn_TYPE_member_t asn_MBR_addresses_3[] = {
	{ ATF_POINTER, 0, 0,
		(ASN_TAG_CLASS_UNIVERSAL | (16 << 2)),
		0,
		&asn_DEF_ROAIPAddress,
		0,
		{ 0, 0, 0 },
		0, 0, /* No default value */
		""
		},
};
static const ber_tlv_tag_t asn_DEF_addresses_tags_3[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static asn_SET_OF_specifics_t asn_SPC_addresses_specs_3 = {
	sizeof(struct ROAIPAddressFamily__addresses),
	offsetof(struct ROAIPAddressFamily__addresses, _asn_ctx),
	0,	/* XER encoding is XMLDelimitedItemList */
};
static /* Use -fall-defs-global to expose */
asn_TYPE_descriptor_t asn_DEF_addresses_3 = {
	"addresses",
	"addresses",
	&asn_OP_SEQUENCE_OF,
	asn_DEF_addresses_tags_3,
	sizeof(asn_DEF_addresses_tags_3)
		/sizeof(asn_DEF_addresses_tags_3[0]), /* 1 */
	asn_DEF_addresses_tags_3,	/* Same as above */
	sizeof(asn_DEF_addresses_tags_3)
		/sizeof(asn_DEF_addresses_tags_3[0]), /* 1 */
	{ &asn_OER_type_addresses_constr_3, 0, SEQUENCE_OF_constraint },
	asn_MBR_addresses_3,
	1,	/* Single element */
	&asn_SPC_addresses_specs_3	/* Additional specs */
};

asn_TYPE_member_t asn_MBR_ROAIPAddressFamily_1[] = {
	{ ATF_NOFLAGS, 0, offsetof(struct ROAIPAddressFamily, addressFamily),
		(ASN_TAG_CLASS_UNIVERSAL | (4 << 2)),
		0,
		&asn_DEF_OCTET_STRING,
		0,
		{ &asn_OER_memb_addressFamily_constr_2, 0,  memb_addressFamily_constraint_1 },
		0, 0, /* No default value */
		"addressFamily"
		},
	{ ATF_NOFLAGS, 0, offsetof(struct ROAIPAddressFamily, addresses),
		(ASN_TAG_CLASS_UNIVERSAL | (16 << 2)),
		0,
		&asn_DEF_addresses_3,
		0,
		{ &asn_OER_memb_addresses_constr_3, 0,  memb_addresses_constraint_1 },
		0, 0, /* No default value */
		"addresses"
		},
};
static const ber_tlv_tag_t asn_DEF_ROAIPAddressFamily_tags_1[] = {
	(ASN_TAG_CLASS_UNIVERSAL | (16 << 2))
};
static const asn_TYPE_tag2member_t asn_MAP_ROAIPAddressFamily_tag2el_1[] = {
    { (ASN_TAG_CLASS_UNIVERSAL | (4 << 2)), 0, 0, 0 }, /* addressFamily */
    { (ASN_TAG_CLASS_UNIVERSAL | (16 << 2)), 1, 0, 0 } /* addresses */
};
asn_SEQUENCE_specifics_t asn_SPC_ROAIPAddressFamily_specs_1 = {
	sizeof(struct ROAIPAddressFamily),
	offsetof(struct ROAIPAddressFamily, _asn_ctx),
	asn_MAP_ROAIPAddressFamily_tag2el_1,
	2,	/* Count of tags in the map */
	0, 0, 0,	/* Optional elements (not needed) */
	-1,	/* First extension addition */
};
asn_TYPE_descriptor_t asn_DEF_ROAIPAddressFamily = {
	"ROAIPAddressFamily",
	"ROAIPAddressFamily",
	&asn_OP_SEQUENCE,
	asn_DEF_ROAIPAddressFamily_tags_1,
	sizeof(asn_DEF_ROAIPAddressFamily_tags_1)
		/sizeof(asn_DEF_ROAIPAddressFamily_tags_1[0]), /* 1 */
	asn_DEF_ROAIPAddressFamily_tags_1,	/* Same as above */
	sizeof(asn_DEF_ROAIPAddressFamily_tags_1)
		/sizeof(asn_DEF_ROAIPAddressFamily_tags_1[0]), /* 1 */
	{ 0, 0, SEQUENCE_constraint },
	asn_MBR_ROAIPAddressFamily_1,
	2,	/* Elements count */
	&asn_SPC_ROAIPAddressFamily_specs_1	/* Additional specs */
};

