/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#include "CMSAttributeValue.h"

/*
 * This type is implemented using ANY,
 * so here we adjust the DEF accordingly.
 */
asn_TYPE_descriptor_t asn_DEF_CMSAttributeValue = {
	"CMSAttributeValue",
	"CMSAttributeValue",
	&asn_OP_ANY,
	0,	/* No effective tags (pointer) */
	0,	/* No effective tags (count) */
	0,	/* No tags (pointer) */
	0,	/* No tags (count) */
	{ 0, 0, ANY_constraint },
	0, 0,	/* No members */
	&asn_SPC_ANY_specs	/* Additional specs */
};

