/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_SignedAttributes_H_
#define	_SignedAttributes_H_


#include "asn1/asn1c/asn_application.h"

/* Including external dependencies */
#include "asn1/asn1c/asn_SET_OF.h"
#include "asn1/asn1c/constr_SET_OF.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
struct CMSAttribute;

/* SignedAttributes */
typedef struct SignedAttributes {
	A_SET_OF(struct CMSAttribute) list;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} SignedAttributes_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_SignedAttributes;
extern asn_SET_OF_specifics_t asn_SPC_SignedAttributes_specs_1;
extern asn_TYPE_member_t asn_MBR_SignedAttributes_1[1];

#ifdef __cplusplus
}
#endif

/* Referred external types */
#include "CMSAttribute.h"

#endif	/* _SignedAttributes_H_ */
#include "asn1/asn1c/asn_internal.h"
