#ifndef SRC_CONFIG_RRDP_CONF_H_
#define SRC_CONFIG_RRDP_CONF_H_

#include "config/types.h"

extern const struct global_type gt_rrdp_enabled;
extern const struct global_type gt_rrdp_priority;
extern const struct global_type gt_rrdp_retry_count;
extern const struct global_type gt_rrdp_retry_interval;

#endif /* SRC_CONFIG_RRDP_CONF_H_ */
