/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "CryptographicMessageSyntax2004"
 * 	found in "rfc5652-12.1.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_SignerIdentifier_H_
#define	_SignerIdentifier_H_


#include "asn1/asn1c/asn_application.h"

/* Including external dependencies */
#include "IssuerAndSerialNumber.h"
#include "CMSSubjectKeyIdentifier.h"
#include "asn1/asn1c/constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum SignerIdentifier_PR {
	SignerIdentifier_PR_NOTHING,	/* No components present */
	SignerIdentifier_PR_issuerAndSerialNumber,
	SignerIdentifier_PR_subjectKeyIdentifier
} SignerIdentifier_PR;

/* SignerIdentifier */
typedef struct SignerIdentifier {
	SignerIdentifier_PR present;
	union SignerIdentifier_u {
		IssuerAndSerialNumber_t	 issuerAndSerialNumber;
		CMSSubjectKeyIdentifier_t	 subjectKeyIdentifier;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} SignerIdentifier_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_SignerIdentifier;
extern asn_CHOICE_specifics_t asn_SPC_SignerIdentifier_specs_1;
extern asn_TYPE_member_t asn_MBR_SignerIdentifier_1[2];

#ifdef __cplusplus
}
#endif

#endif	/* _SignerIdentifier_H_ */
#include "asn1/asn1c/asn_internal.h"
