#ifndef SRC_OBJECT_GHOSTBUSTERS_H_
#define SRC_OBJECT_GHOSTBUSTERS_H_

#include "rpp.h"
#include "types/uri.h"

int ghostbusters_traverse(struct rpki_uri *, struct rpp *);

#endif /* SRC_OBJECT_GHOSTBUSTERS_H_ */
