/*
 * Generated by asn1c-0.9.29 (http://lionet.info/asn1c)
 * From ASN.1 module "SignedObjectSyntax"
 * 	found in "rfc6488.asn1"
 * 	`asn1c -Werror -fcompound-names -fwide-types -D asn1/asn1c -no-gen-PER -no-gen-example`
 */

#ifndef	_BinarySigningTime_H_
#define	_BinarySigningTime_H_


#include "asn1/asn1c/asn_application.h"

/* Including external dependencies */
#include "BinaryTime.h"

#ifdef __cplusplus
extern "C" {
#endif

/* BinarySigningTime */
typedef BinaryTime_t	 BinarySigningTime_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_BinarySigningTime;
asn_struct_free_f BinarySigningTime_free;
asn_struct_print_f BinarySigningTime_print;
asn_constr_check_f BinarySigningTime_constraint;
ber_type_decoder_f BinarySigningTime_decode_ber;
der_type_encoder_f BinarySigningTime_encode_der;
xer_type_decoder_f BinarySigningTime_decode_xer;
xer_type_encoder_f BinarySigningTime_encode_xer;
oer_type_decoder_f BinarySigningTime_decode_oer;
oer_type_encoder_f BinarySigningTime_encode_oer;

#ifdef __cplusplus
}
#endif

#endif	/* _BinarySigningTime_H_ */
#include "asn1/asn1c/asn_internal.h"
